from series.db import FileDatabase as Base
from series.library.model.episode import Episode


class FileDatabase(Base):

    def __init__(self, *a, **kw):
        super().__init__('series.library', *a, **kw)
        if self._connected and not self._outdated:
            self._sanity_check()

    def _sanity_check(self):
        for epi in self.query(Episode):
            if epi.series is None:
                self.session.delete(epi)
        self.commit()

__all__ = ['FileDatabase']



from golgi.config import PathConfigOption, ListConfigOption

metadata = dict(parents=['tek_utils.sharehoster'])


def reset_config():
    return {
        'series': dict(
            series_dir=PathConfigOption(),
            temp_dir=PathConfigOption(),
            monitor=ListConfigOption(),
            library_url='http://localhost:8111',
            enumeration_regex='',
        ),
        'client': dict(cli_cmd=None, cli_params=[],)
    }



from datetime import datetime

import requests

from golgi.config import configurable

from series.get.handler import ReleaseHandler


# TODO use ApiClient
@configurable(series=['library_url'], get=['library'])
class LibraryHandler(ReleaseHandler):
    _library_path = '/series/{}/season/{}/episode/{}'

    def __init__(self, releases, shows, *a, **kw):
        super().__init__(releases, 5, 'library handler', **kw)
        self._last_error = None
        self._error_interval = 7200
        self._shows = shows

    def _qualify(self, monitor):
        return not monitor.added_to_library and monitor.downloaded

    def _handle(self, monitor):
        if self._library:
            release = monitor.release
            show = self._shows.find_by_name(release.name)
            tvdb_id = show.map(lambda a: a.etvdb_id).get_or_strict(None)
            try:
                requests.post(self._library_url +
                              self._library_path.format(release.name,
                                                        release.season,
                                                        release.episode), json=dict(tvdb_id=tvdb_id))
            except requests.RequestException:
                self._connect_error()
            else:
                self._update(monitor, added_to_library=True)

    def _connect_error(self):
        if (self._last_error is None or
                (datetime.now() - self._last_error).total_seconds() >
                self._error_interval):
            self.log.error("Couldn't connect to library host!")
            self._last_error = datetime.now()


__all__ = ['LibraryHandler']



# coding: utf-8

from golgi import configurable
from tek.errors import ParseError

from amino import IO, do, Do

from series.library.handler import BaseHandler
from series.library.model.season import Season
from series.library.model.series import Series
from series.tvdb.data import TvdbEpisodeMetadata
from series.tvdb.api import guess_show, season_metadata, episode_metadata


@configurable(library=['metadata_interval'])
class Metadata(BaseHandler):

    def __init__(self, library, player):
        super().__init__(library, self._metadata_interval, 'metadata service')
        self._library = library
        self._player = player
        self._min_failed = 0

    def _handle(self, item):
        data = {}
        handler = (self._handle_season if isinstance(item, Season) else self._handle_episode)
        try:
            self.log.info('Fetching metadata for {}…'.format(item))
            handler(item).fatal
        except ParseError as e:
            self.log.error(e)
            data = dict(metadata_failures=item.metadata_failures + 1)
        else:
            data = dict(metadata_fetched=True)
        self._library.alter_object(item, data)

    @do(IO[None])
    def _handle_season(self, season) -> Do:
        if not season.series.tvdb_id:
            yield self.update_tvdb_id(season.series)
        if season.series.tvdb_id:
            data = yield season_metadata(season.series.tvdb_id, season.number)
            data.map(lambda epis: epis.map(lambda epi: self._update_episode(season.series, epi)))


    @do(IO[None])
    def _handle_episode(self, episode) -> Do:
        if not episode.series.tvdb_id:
            yield self.update_tvdb_id(episode.series)
        if episode.series.tvdb_id:
            data = yield episode_metadata(episode.series.tvdb_id,
                                          episode.season.number,
                                          episode.number)
            data.map(lambda epi: self._update_episode(episode.season.series, epi))

    @do(IO[None])
    def update_tvdb_id(self, series: Series) -> Do:
        response = yield guess_show(series.name)
        may_show = yield IO.from_either(response)
        show = yield IO.from_maybe(may_show, lambda: f'could not find a tvdb show for {series.name}')
        yield IO.delay(self._library.alter_series, series.name, dict(tvdb_id=show.id))

    def _update_episode(self, series, data: TvdbEpisodeMetadata):
        self._library.alter_episode(
            series,
            data.season,
            data.number,
            dict(title=data.title, overview=data.overview, metadata_fetched=True),
        )

    def _qualify(self, item):
        fails = item.metadata_failures
        return fails is not None and fails == self._min_failed and fails < 15

    @property
    def _candidates(self):
        candidates = (self._seasons or self._new_episodes or
                      self._failed_episodes)
        counts = [c.metadata_failures for c in candidates
                  if c.metadata_failures is not None]
        self._min_failed = min(counts or [0])
        return candidates

    @property
    def _new_episodes(self):
        return [e for e in self._library.new_episodes if not
                e.metadata_fetched]

    @property
    def _seasons(self):
        return self._library.seasons(extra=dict(metadata_fetched=False))

    @property
    def _failed_episodes(self):
        return []


__all__ = ['Metadata']



from multiprocessing import Process, Array, Value
from datetime import datetime

from golgi.config import configurable
from amino import Maybe
from amino.lazy_list import lazy_list

from series.get.handler import ReleaseHandler
from series.get.model.link import LinkChecker
from series.logging import Logging
from series.handler import Job


@configurable(get=['link_check_proc_timeout'])
class LinkCheckTask(Process, Logging):

    def __init__(self, link):
        self._link_check_proc_timeout
        self.link = link
        self.done = Value('i', 0)
        self.status = Value('i', 0)
        self.dead = Value('i', 0)
        self.reason = Array('c', 1024)
        self.size = Value('i', 0)
        self.multipart = Value('i', 0)
        self.link.last_check = datetime.now()
        super().__init__(name='seriesd link checker')
        self.get = True

    def run(self):
        self.log.debug('Checking link {}…'.format(self.link.url))
        self._check_link()
        self.done.value = 1

    @property
    def timeout(self):
        delta = (datetime.now() - self.link.last_check).total_seconds()
        return delta > self._link_check_proc_timeout

    def _check_link(self):
        checker = LinkChecker(self.link)
        status, dead, reason, size, multipart = checker.result
        self.status.value = status
        self.dead.value = dead
        self.reason.value = reason.encode('utf8')[:1024]
        self.size.value = size
        self.multipart.value = multipart

    @property
    def result(self):
        return [self.status.value, self.dead.value, self.reason.value.decode(),
                self.size.value, self.multipart.value]


@configurable(get=['link_check_procs'])
class LinkHandler(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        self._running_tasks = []
        super().__init__(releases, 1, 'link handler', **kw)

    def _handle(self, links):
        for link in links:
            if len(self._running_tasks) < self._link_check_procs:
                self._launch_task(link)
        self._cleanup()

    @property
    def _current(self):
        return Maybe.check(Job(self._candidates))

    def _launch_task(self, link):
        link.checking = True
        task = LinkCheckTask(link)
        self._running_tasks.append(task)
        task.start()

    def _cleanup(self):
        self._clean_done()
        self._clean_timeouts()

    def _clean_done(self):
        done = [task for task in self._running_tasks if task.done.value]
        for task in done:
            task.link.checking = False
            self._running_tasks.remove(task)
            task.link.set_status(*task.result)
        if done:
            self._commit()

    def _clean_timeouts(self):
        timeout = [task for task in self._running_tasks if task.timeout]
        for task in timeout:
            task.terminate()
            task.link.checking = False
            task.link.check_failed()
            self._running_tasks.remove(task)
        if timeout:
            self._commit()

    def wait(self):
        while self._running_tasks:
            self._cleanup()
        self._cleanup()

    def terminate(self):
        for task in self._running_tasks:
            task.terminate()

    def stop(self):
        super().stop()
        self.terminate()

    @property
    def _candidates(self):
        return self._critical or self._normal or self._recheck

    @property
    def _critical(self):
        return self._pick_links(self._critical_releases)

    @property
    def _normal(self):
        return self._pick_links(self._normal_releases)

    @property
    def _recheck(self):
        return self._pick_links(self._normal_releases, recheck=True)

    @lazy_list
    def _pick_links(self, releases, recheck=False):
        attr = ('re' if recheck else '') + 'checkable_links'
        for release in releases:
            links = getattr(release, attr)
            if links:
                yield links[0]

    @property
    def _critical_releases(self):
        return [r for r in self._normal_releases if not r.valid_links]

    @property
    def _normal_releases(self):
        return self._releases.download_candidates

__all__ = ['LinkHandler']



from tek_utils.sharehoster.torrent import _clients, _cachers

downloadable = {}  # type: dict
caching = {}  # type: dict


class SpecCacher(object):

    def __init__(self, torrent=None, file_id=None, largest_file=True):
        self.torrent = torrent
        self._file_id = file_id
        self._largest_file = largest_file
        self._create_info = {}
        self._info = {}
        self.downloaded = downloadable.get(self.torrent, True)

    @property
    def caching(self):
        return caching.get(self.torrent, False)

    def request(self):
        pass

    @property
    def download_url(self):
        return 'http://host/{}/filename'.format(self.torrent[7:])

    @property
    def status(self):
        return ('IN_QUEUE' if self.caching else 'DOWNLOADED' if self.downloaded
                else 'NONE')

    @property
    def status_message(self):
        return ('up: 5 down: 2' if self.caching else 'downloaded' if
                self.downloaded else 'nope')


class SpecClient(object):

    @property
    def account_info(self):
        return True

    @property
    def transfers(self):
        return []

_clients['spec'] = SpecClient
_cachers['spec'] = SpecCacher



from tek.errors import TException

class UnknownSeries(TException):
    def __init__(self, name, best, best_ratio):
        text = 'Unknown series "{}"! (Best match was {} with ratio {})'
        super().__init__(text.format(name, best, best_ratio))



from mplayer import Player

from series.library.player.interface import PlayerInterface


class MPlayer(PlayerInterface):

    def __init__(self, args):
        def dict2args(a):
            return ["-{}={}".format(k, v) for k, v in a.items()]
        base_args = ('-slave', '-idle', '-really-quiet', '-msglevel',
                     'global=4', '-noconfig', 'all')
        super().__init__(Player(args=dict2args(args) + base_args))

    def stop(self):
        self._player.stop()

    def pause(self):
        self._player.pause()

    def osd(self, level):
        self._player.osd(level)

    def osd_level(self):
        return self._player.osdlevel

    @property
    def current_volume(self):
        return self._player.volume

    def volume(self, value):
        self._player.volume = value

    def message(self, msg, duration):
        pass

__all__ = ['MPlayer']



import os
import glob
import subprocess
import shlex
import shutil

from golgi.config import configurable
from tek.tools import first_valid
from series.logging import Logging

from series import is_video


class Collection(Logging):

    def __init__(self, path):
        self._path = path

    @property
    def video_files(self):
        for dirpath, _, files in os.walk(str(self._path)):
            for _file in files:
                if is_video(_file):
                    yield '/'.join((dirpath, _file))

    def contains_video(self, video):
        return self.video_path(video) is not None


@configurable(library=['path_template', 'subtitle_path_template',
                       'name_formatter'])
class EpisodeCollection(Collection):
    ''' Represents a single directory tree containing video and subtitle
    files.
    Provides helpers for iterating all containing files and assembling
    paths to episodes and subtitles.
    '''

    def video_path(self, episode):
        return self._file_path(self._path_template, episode)

    def subtitle_path(self, episode):
        return self._file_path(self._subtitle_path_template, episode)

    def _file_path(self, template, episode):
        relative_path = template.format(
            name=episode.series.canonical_name,
            formatted=self._format_name(episode.series.canonical_name),
            season=episode.season.number,
            episode=episode.number,
            ext='*',
        )
        return first_valid(self._path.glob(relative_path))

    def _format_name(self, name):
        if self._name_formatter:
            parts = shlex.split(self._name_formatter.format(name))
            if shutil.which(parts[0]):
                try:
                    name = subprocess.check_output(parts).decode().strip()
                except subprocess.CalledProcessError as e:
                    self.log.error('External name formatter failed!')
                    self.log.error(e)
        return name


@configurable(library=['movie_path_template',
                       'movie_subtitle_path_template'])
class MovieCollection(Collection):

    def video_path(self, movie):
        return self._file_path(self._movie_path_template, movie)

    def subtitle_path(self, movie):
        return self._file_path(self._movie_subtitle_path_template, movie)

    def _file_path(self, template, movie):
        relative_path = template.format(title=movie.title, ext='*')
        return first_valid(self._path.glob(relative_path))

__all__ = ['EpisodeCollection', 'MovieCollection']
